// Linear actuator driver
// May 9, 2020
// Written by Theron Wierenga, theron@wierengaconsulting.com
// Two L298N motor driver modules run by Arduino with eight Hall A3144 sensors
// Working 6/11/2020

#define HALL1 A5
#define HALL2 A4
#define HALL3 2
#define HALL4 4
#define HALL5 12
#define HALL6 13
#define HALL7 A7
#define HALL8 A6

#define HALL_ON  0
#define HALL_OFF 1

#define ENABLE_A  9
#define IN_A1     8
#define IN_A2     7

#define ENABLE_B  3
#define IN_B1     6
#define IN_B2     5

#define ENABLE_C  11
#define IN_C1     A0
#define IN_C2     A1

#define ENABLE_D  10
#define IN_D1     A2
#define IN_D2     A3

#define MOTOR_A 0
#define MOTOR_B 1
#define MOTOR_C 2
#define MOTOR_D 3

#define MOTOR_OFF     0
#define MOTOR_BRAKED  1
#define MOTOR_IN      2
#define MOTOR_OUT     3

long t1, t2, t3, t4;
void setup()
{
    delay(4000);
    Serial.begin(115200);
    Serial.println("Starting");
    pinMode(HALL1, INPUT_PULLUP);
    pinMode(HALL2, INPUT_PULLUP);
    pinMode(HALL3, INPUT_PULLUP);
    pinMode(HALL4, INPUT_PULLUP);
    pinMode(HALL5, INPUT_PULLUP);
    pinMode(HALL6, INPUT_PULLUP);
    pinMode(HALL7, INPUT_PULLUP);
    pinMode(HALL8, INPUT_PULLUP);

    pinMode(ENABLE_A, OUTPUT);
    pinMode(IN_A1, OUTPUT);
    pinMode(IN_A2, OUTPUT);
    
    pinMode(ENABLE_B, OUTPUT);
    pinMode(IN_B1, OUTPUT);
    pinMode(IN_B2, OUTPUT);
    
    pinMode(ENABLE_C, OUTPUT);
    pinMode(IN_C1, OUTPUT);
    pinMode(IN_C2, OUTPUT);
    
    pinMode(ENABLE_D, OUTPUT);
    pinMode(IN_D1, OUTPUT);
    pinMode(IN_D2, OUTPUT);

    Serial.println("Home Motor");
    homeMotor(MOTOR_A, HALL1, 255);
    delay(500);
}

void loop()
{
    // Run motor A out and in
    // Assume Hall sensor at home
    softStartPWM(MOTOR_A, MOTOR_OUT, 255);
    // Move off the Hall sensor
    while (readHall(HALL1) == HALL_ON) {};
    // Wait for Hall sensor to go back on
    while (readHall(HALL1) == HALL_OFF) {};
    // Stop motor
    setMotor(MOTOR_A, MOTOR_BRAKED);
    delay(200);
    // Reverse motor
    softStartPWM(MOTOR_A, MOTOR_IN, 255);
    // Move off Hall sensor
    while (readHall(HALL1) == HALL_ON) {}
    // Wait for Hall sensor to go back on
    while (readHall(HALL1) == HALL_OFF) {};
    // Stop motor
    setMotor(MOTOR_A, MOTOR_BRAKED);   
    delay(200);
  
    
}

// Soft start by increasing PWM before full power
bool softStartPWM(int motor, int setting, int maxPWM)
{
    if ((motor != MOTOR_A) && (motor != MOTOR_B) && (motor != MOTOR_C) && (motor != MOTOR_D)) return false;
    if ((setting != MOTOR_IN) && (setting != MOTOR_OUT)) return false;
    if (!((maxPWM >= 0) && (maxPWM <= 255))) return false;
    switch (motor)
    {
      case MOTOR_A:switch(setting)
                   {
                     case MOTOR_IN:  digitalWrite(IN_A1, LOW);
                                     digitalWrite(IN_A2, HIGH);
                                     break;
                     case MOTOR_OUT: digitalWrite(IN_A1, HIGH);
                                     digitalWrite(IN_A2, LOW);
                                     break;                   
                   }
                   for (int i = 32; i <= maxPWM; i++)
                   {
                      analogWrite(ENABLE_A, i);
                      delay(1);
                   }
                   break;
      case MOTOR_B:switch(setting)
                   {
                     case MOTOR_IN:  digitalWrite(IN_B1, LOW);
                                     digitalWrite(IN_B2, HIGH);
                                     break;
                     case MOTOR_OUT: digitalWrite(IN_B1, HIGH);
                                     digitalWrite(IN_B2, LOW);
                                     break;                   
                   }
                   for (int i = 32; i <= maxPWM; i++)
                   {
                      analogWrite(ENABLE_B, i);
                      delay(1);
                   }
                   break;
      case MOTOR_C:switch(setting)
                   {
                     case MOTOR_IN:  digitalWrite(IN_C1, LOW);
                                     digitalWrite(IN_C2, HIGH);
                                     break;
                     case MOTOR_OUT: digitalWrite(IN_C1, HIGH);
                                     digitalWrite(IN_C2, LOW);
                                     break;                   
                   }
                   for (int i = 32; i <= maxPWM; i++)
                   {
                      analogWrite(ENABLE_C, i);
                      delay(1);
                   }
                   break;
      case MOTOR_D:switch(setting)
                   {
                     case MOTOR_IN:  digitalWrite(IN_D1, LOW);
                                     digitalWrite(IN_D2, HIGH);
                                     break;
                     case MOTOR_OUT: digitalWrite(IN_D1, HIGH);
                                     digitalWrite(IN_D2, LOW);
                                     break;                   
                   }
                   for (int i = 32; i <= maxPWM; i++)
                   {
                      analogWrite(ENABLE_D, i);
                      delay(1);
                   }
                   break;
    }
    return true;
}

// This function is used to turn on a motor, running IN or OUT, using PWM
bool setMotorPWM(int motor, int setting, int pwm)
{
    if ((motor != MOTOR_A) && (motor != MOTOR_B) && (motor != MOTOR_C) && (motor != MOTOR_D)) return false;
    if ((setting != MOTOR_IN) && (setting != MOTOR_OUT)) return false;
    if (!((pwm >= 0) && (pwm <= 255))) return false;
    switch (motor)
    {
      case MOTOR_A:switch(setting)
                   {
                     case MOTOR_IN:  digitalWrite(IN_A1, LOW);
                                     digitalWrite(IN_A2, HIGH);
                                     break;
                     case MOTOR_OUT: digitalWrite(IN_A1, HIGH);
                                     digitalWrite(IN_A2, LOW);
                                     break;                   
                   }
                   analogWrite(ENABLE_A, pwm);
                   break;
      case MOTOR_B:switch(setting)
                   {
                     case MOTOR_IN:  digitalWrite(IN_B1, LOW);
                                     digitalWrite(IN_B2, HIGH);
                                     break;
                     case MOTOR_OUT: digitalWrite(IN_B1, HIGH);
                                     digitalWrite(IN_B2, LOW);
                                     break;                   
                   }
                   analogWrite(ENABLE_B, pwm);
                   break;
      case MOTOR_C:switch(setting)
                   {
                     case MOTOR_IN:  digitalWrite(IN_C1, LOW);
                                     digitalWrite(IN_C2, HIGH);
                                     break;
                     case MOTOR_OUT: digitalWrite(IN_C1, HIGH);
                                     digitalWrite(IN_C2, LOW);
                                     break;                   
                   }
                   analogWrite(ENABLE_C, pwm);
                   break;
      case MOTOR_D:switch(setting)
                   {
                     case MOTOR_IN:  digitalWrite(IN_D1, LOW);
                                     digitalWrite(IN_D2, HIGH);
                                     break;
                     case MOTOR_OUT: digitalWrite(IN_D1, HIGH);
                                     digitalWrite(IN_D2, LOW);
                                     break;                   
                   }
                   analogWrite(ENABLE_D, pwm);
                   break;
    }
    return true;
}

// This function is used to turn a motor ON, OFF, BRAKED or OFF (free turning) at full power
bool setMotor(int motor, int setting)
{
    if ((motor != MOTOR_A) && (motor != MOTOR_B) && (motor != MOTOR_C) && (motor != MOTOR_D)) return false;
    if ((setting != MOTOR_OFF) && (setting != MOTOR_BRAKED) && (setting != MOTOR_IN) && (setting != MOTOR_OUT)) return false;
    switch (motor)
    {
      case MOTOR_A:switch (setting)
                   {
                        case MOTOR_OFF:    digitalWrite(ENABLE_A, LOW);
                                           break;
                        case MOTOR_BRAKED: digitalWrite(IN_A1, LOW);
                                           digitalWrite(IN_A2, LOW);
                                           digitalWrite(ENABLE_A, HIGH);
                                           break;
                        case MOTOR_IN:     digitalWrite(IN_A1, LOW);
                                           digitalWrite(IN_A2, HIGH);
                                           digitalWrite(ENABLE_A, HIGH);
                                           break;
                        case MOTOR_OUT:    digitalWrite(IN_A1, HIGH);
                                           digitalWrite(IN_A2, LOW);
                                           digitalWrite(ENABLE_A, HIGH);
                                           break;
                   }
                   break;
      case MOTOR_B:switch (setting)
                   {
                        case MOTOR_OFF:    digitalWrite(ENABLE_B, LOW);
                                           break;
                        case MOTOR_BRAKED: digitalWrite(IN_B1, LOW);
                                           digitalWrite(IN_B2, LOW);
                                           digitalWrite(ENABLE_B, HIGH);
                                           break;
                        case MOTOR_IN:     digitalWrite(IN_B1, LOW);
                                           digitalWrite(IN_B2, HIGH);
                                           digitalWrite(ENABLE_B, HIGH);
                                           break;
                        case MOTOR_OUT:    digitalWrite(IN_B1, HIGH);
                                           digitalWrite(IN_B2, LOW);
                                           digitalWrite(ENABLE_B, HIGH);
                                           break;
                   }
                   break;
      case MOTOR_C:switch (setting)
                   {
                        case MOTOR_OFF:    digitalWrite(ENABLE_C, LOW);
                                           break;
                        case MOTOR_BRAKED: digitalWrite(IN_C1, LOW);
                                           digitalWrite(IN_C2, LOW);
                                           digitalWrite(ENABLE_C, HIGH);
                                           break;
                        case MOTOR_IN:     digitalWrite(IN_C1, LOW);
                                           digitalWrite(IN_C2, HIGH);
                                           digitalWrite(ENABLE_C, HIGH);
                                           break;
                        case MOTOR_OUT:    digitalWrite(IN_C1, HIGH);
                                           digitalWrite(IN_C2, LOW);
                                           digitalWrite(ENABLE_C, HIGH);
                                           break;
                   }
                   break;
      case MOTOR_D:switch (setting)
                   {
                        case MOTOR_OFF:    digitalWrite(ENABLE_D, LOW);
                                           break;
                        case MOTOR_BRAKED: digitalWrite(IN_D1, LOW);
                                           digitalWrite(IN_D2, LOW);
                                           digitalWrite(ENABLE_D, HIGH);
                                           break;
                        case MOTOR_IN:     digitalWrite(IN_D1, LOW);
                                           digitalWrite(IN_D2, HIGH);
                                           digitalWrite(ENABLE_D, HIGH);
                                           break;
                        case MOTOR_OUT:    digitalWrite(IN_D1, HIGH);
                                           digitalWrite(IN_D2, LOW);
                                           digitalWrite(ENABLE_D, HIGH);
                                           break;
                   }
                   break;
    }
    return true;
}

// Read the Hall sensor
int readHall(int num)
{
   return digitalRead(num); 
}

// Return motor to home positon, which is fully retracted
bool homeMotor(int motor, int hall, int pwm)
{
    if ((motor != MOTOR_A) && (motor != MOTOR_B) && (motor != MOTOR_C) && (motor != MOTOR_D)) return false;
    setMotorPWM(motor, MOTOR_IN, pwm);
    while (readHall(hall) == HALL_OFF) {};
    setMotor(motor, MOTOR_BRAKED);
    return true;
}

// Speed check
// Start at home position, runs out for 1/4 second
void speedCheck()
{
    long t1 = millis();
    setMotor(MOTOR_A, MOTOR_OUT);
    while ((millis() - t1) < 250) {};
    setMotor(MOTOR_A, MOTOR_BRAKED);
}
